@@ -24,7 +24,8 @@ cudaError_t cudaGetLastError() {
 
 
 // Initialization
-cudaError_t cudaDriverGetVersion(...) {
+cudaError_t cudaDriverGetVersion(int* driverVersion) {
+    *driverVersion = 0;  // make the stub work safely without driver
     return cudaSuccess;
 }
 
@@ -74,6 +75,10 @@ cudaError_t cudaDeviceEnablePeerAccess(...) {
     return cudaSuccess;
 }
 
+cudaError_t cudaDeviceDisablePeerAccess(...) {
+    return cudaSuccess;
+}
+
 cudaError_t cudaDeviceGetLimit(...) {
     return cudaSuccess;
 }
@@ -240,6 +245,14 @@ cudaError_t cudaMemPoolTrimTo(...) {
     return cudaSuccess;
 }
 
+cudaError_t cudaMemPoolGetAttribute(...) {
+    return cudaSuccess;
+}
+
+cudaError_t cudaMemPoolSetAttribute(...) {
+    return cudaSuccess;
+}
+
 
 // Stream and Event
 cudaError_t cudaStreamCreate(...) {
