@@ -1,20 +1,22 @@
-import numpy
+import string
+import warnings
 
+import numpy
 try:
     import scipy.sparse
     scipy_available = True
 except ImportError:
     scipy_available = False
 
-import warnings
-
 import cupy
 import cupyx
 
-from cupy import core
+from cupy import _core
+from cupy._core import _scalar
 from cupy._creation import basic
 from cupy import cusparse
 from cupyx.scipy.sparse import base
+from cupyx.scipy.sparse import coo
 from cupyx.scipy.sparse import data as sparse_data
 from cupyx.scipy.sparse import sputils
 from cupyx.scipy.sparse import _util
@@ -26,9 +28,9 @@ class _compressed_sparse_matrix(sparse_data._data_matrix,
                                 sparse_data._minmax_mixin,
                                 _index.IndexMixin):
 
-    _max_reduction_kern = core.RawKernel(r'''
+    _max_min_reduction_code = r'''
         extern "C" __global__
-        void max_reduction(double* data, int* x, int* y, int length,
+        void ${func}(double* data, int* x, int* y, int length,
                            double* z) {
             // Get the index of the block
             int tid = blockIdx.x * blockDim.x + threadIdx.x;
@@ -38,7 +40,7 @@ class _compressed_sparse_matrix(sparse_data._data_matrix,
 
             // Select initial value based on the block density
             double running_value = 0;
-            if (block_length == length){
+            if (${cond}){
                 running_value = data[x[tid]];
             } else {
                 running_value = 0;
@@ -52,7 +54,7 @@ class _compressed_sparse_matrix(sparse_data._data_matrix,
                     break;
                 } else {
                     // Check for a value update
-                    if (data[entry] > running_value){
+                    if (data[entry] ${op} running_value){
                         running_value = data[entry];
                     }
                 }
@@ -60,185 +62,42 @@ class _compressed_sparse_matrix(sparse_data._data_matrix,
 
             // Store in the return function
             z[tid] = running_value;
-        }
-        ''', 'max_reduction')
+        }'''
 
-    _max_nonzero_reduction_kern = core.RawKernel(r'''
-        extern "C" __global__
-        void max_nonzero_reduction(double* data, int* x, int* y, int length,
-                                   double* z) {
+    _max_reduction_kern = _core.RawKernel(
+        string.Template(_max_min_reduction_code).substitute(
+            func='max_reduction', op='>', cond='block_length == length'),
+        'max_reduction')
+
+    _max_nonzero_reduction_kern = _core.RawKernel(
+        string.Template(_max_min_reduction_code).substitute(
+            func='max_nonzero_reduction', op='>', cond='block_length > 0'),
+        'max_nonzero_reduction')
+
+    _min_reduction_kern = _core.RawKernel(
+        string.Template(_max_min_reduction_code).substitute(
+            func='min_reduction', op='<', cond='block_length == length'),
+        'min_reduction')
+
+    _min_nonzero_reduction_kern = _core.RawKernel(
+        string.Template(_max_min_reduction_code).substitute(
+            func='min_nonzero_reduction', op='<', cond='block_length > 0'),
+        'min_nonzero_reduction')
+
+    # For _max_arg_reduction_mod and _min_arg_reduction_mod below, we pick
+    # the right template specialization according to input dtypes at runtime.
+    # The distinction in int types (T2) is important for portability in OS.
+
+    _argmax_argmin_code = r'''
+        template<typename T1, typename T2> __global__ void
+        ${func}_arg_reduction(T1* data, int* indices, int* x, int* y,
+                              int length, T2* z) {
             // Get the index of the block
             int tid = blockIdx.x * blockDim.x + threadIdx.x;
 
             // Calculate the block length
             int block_length = y[tid] - x[tid];
 
-            // Select initial value based on the block density
-            double running_value = 0;
-            if (block_length > 0){
-                running_value = data[x[tid]];
-            } else {
-                running_value = 0;
-            }
-
-            // Iterate over the section of the sparse matrix
-            for (int entry = x[tid]; entry < y[tid]; entry++){
-                if (data[entry] != data[entry]){
-                    // Check for NaN
-                    running_value = nan("");
-                    break;
-                } else {
-                    // Check for a value update
-                    if (running_value < data[entry]){
-                        running_value = data[entry];
-                    }
-                }
-            }
-
-            // Store in the return function
-            z[tid] = running_value;
-        }
-        ''', 'max_nonzero_reduction')
-
-    _min_reduction_kern = core.RawKernel(r'''
-        extern "C" __global__
-        void min_reduction(double* data, int* x, int* y, int length,
-                           double* z) {
-            // Get the index of the block
-            int tid = blockIdx.x * blockDim.x + threadIdx.x;
-
-            // Calculate the block length
-            int block_length = y[tid] - x[tid];
-
-            // Select initial value based on the block density
-            double running_value = 0;
-            if (block_length == length){
-                running_value = data[x[tid]];
-            } else {
-                running_value = 0;
-            }
-
-            // Iterate over the block to update the initial value
-            for (int entry = x[tid]; entry < y[tid]; entry++){
-                if (data[entry] != data[entry]){
-                    // Check for NaN
-                    running_value = nan("");
-                    break;
-                } else {
-                    // Check for a value update
-                    if (data[entry] < running_value){
-                        running_value = data[entry];
-                    }
-                }
-            }
-
-            // Store in the return function
-            z[tid] = running_value;
-        }
-        ''', 'min_reduction')
-
-    _min_nonzero_reduction_kern = core.RawKernel(r'''
-        extern "C" __global__
-        void min_nonzero_reduction(double* data, int* x, int* y, int length,
-                                   double* z) {
-            // Get the index of hte block
-            int tid = blockIdx.x * blockDim.x + threadIdx.x;
-
-            // Calculate the block length
-            int block_length = y[tid] - x[tid];
-
-            // Select initial value based on the block density
-            double running_value = 0;
-            if (block_length > 0){
-                running_value = data[x[tid]];
-            } else {
-                running_value = 0;
-            }
-
-            // Iterate over the section of the sparse matrix
-            for (int entry = x[tid]; entry < y[tid]; entry++){
-                if (data[entry] != data[entry]){
-                    // Check for NaN
-                    running_value = nan("");
-                    break;
-                } else {
-                    // Check for a value update
-                    if (running_value > data[entry]){
-                        running_value = data[entry];
-                    }
-                }
-            }
-
-            // Store in the return function
-            z[tid] = running_value;
-        }
-        ''', 'min_nonzero_reduction')
-
-    _max_arg_reduction_kern = core.RawKernel(r'''
-        extern "C" __global__
-        void max_arg_reduction(double* data, int* indices, int* x, int* y,
-                               int length, long long* z) {
-            // Get the index of the block
-            int tid = blockIdx.x * blockDim.x + threadIdx.x;
-
-            // Calculate the block length
-            int block_length = y[tid] - x[tid];
-
-            // Select initial value based on the block density
-            int data_index = 0;
-            double data_value = 0;
-
-            if (block_length == length){
-                // Block is dense. Fill the first value
-                data_value = data[x[tid]];
-                data_index = indices[x[tid]];
-            } else if (block_length > 0)  {
-                // Block has at least one zero. Assign first occurrence as the
-                // starting reference
-                data_value = 0;
-                for (data_index = 0; data_index < length; data_index++){
-                    if (data_index != indices[x[tid] + data_index] ||
-                        x[tid] + data_index >= y[tid]){
-                        break;
-                    }
-                }
-            } else {
-                 // Zero valued array
-                data_value = 0;
-                data_index = 0;
-            }
-
-            // Iterate over the section of the sparse matrix
-            for (int entry = x[tid]; entry < y[tid]; entry++){
-                if (data[entry] != data[entry]){
-                    // Check for NaN
-                    data_value = nan("");
-                    data_index = 0;
-                    break;
-                } else {
-                    // Check for a value update
-                    if (data[entry] > data_value){
-                        data_index = indices[entry];
-                        data_value = data[entry];
-                    }
-                }
-            }
-
-            // Store in the return function
-            z[tid] = data_index;
-        }
-        ''', 'max_arg_reduction')
-
-    _min_arg_reduction_kern = core.RawKernel(r'''
-        extern "C" __global__
-        void min_arg_reduction(double* data, int* indices, int* x, int* y,
-                               int length, long long* z) {
-            // Get the index of hte block
-            int tid = blockIdx.x * blockDim.x + threadIdx.x;
-
-            // Calculate the block length
-            int block_length = y[tid] - x[tid];
-
             // Select initial value based on the block density
             int data_index = 0;
             double data_value = 0;
@@ -272,7 +131,7 @@ class _compressed_sparse_matrix(sparse_data._data_matrix,
                     break;
                 } else {
                     // Check for a value update
-                    if (data[entry] < data_value){
+                    if (data[entry] ${op} data_value){
                         data_index = indices[entry];
                         data_value = data[entry];
                     }
@@ -281,12 +140,28 @@ class _compressed_sparse_matrix(sparse_data._data_matrix,
 
             // Store in the return function
             z[tid] = data_index;
+        }'''
 
-        }
-        ''', 'min_arg_reduction')
+    _max_arg_reduction_mod = _core.RawModule(
+        code=string.Template(_argmax_argmin_code).substitute(
+            func='max', op='>'),
+        options=('-std=c++11',),
+        name_expressions=['max_arg_reduction<float, int>',
+                          'max_arg_reduction<float, long long>',
+                          'max_arg_reduction<double, int>',
+                          'max_arg_reduction<double, long long>'])
+
+    _min_arg_reduction_mod = _core.RawModule(
+        code=string.Template(_argmax_argmin_code).substitute(
+            func='min', op='<'),
+        options=('-std=c++11',),
+        name_expressions=['min_arg_reduction<float, int>',
+                          'min_arg_reduction<float, long long>',
+                          'min_arg_reduction<double, int>',
+                          'min_arg_reduction<double, long long>'])
 
     # TODO(leofang): rewrite a more load-balanced approach than this naive one?
-    _has_sorted_indices_kern = core.ElementwiseKernel(
+    _has_sorted_indices_kern = _core.ElementwiseKernel(
         'raw T indptr, raw T indices',
         'bool diff',
         '''
@@ -297,10 +172,10 @@ class _compressed_sparse_matrix(sparse_data._data_matrix,
             }
         }
         diff = diff_out;
-        ''', 'has_sorted_indices')
+        ''', 'cupyx_scipy_sparse_has_sorted_indices')
 
     # TODO(leofang): rewrite a more load-balanced approach than this naive one?
-    _has_canonical_format_kern = core.ElementwiseKernel(
+    _has_canonical_format_kern = _core.ElementwiseKernel(
         'raw T indptr, raw T indices',
         'bool diff',
         '''
@@ -315,7 +190,7 @@ class _compressed_sparse_matrix(sparse_data._data_matrix,
             }
         }
         diff = diff_out;
-        ''', 'has_canonical_format')
+        ''', 'cupyx_scipy_sparse_has_canonical_format')
 
     def __init__(self, arg1, shape=None, dtype=None, copy=False):
         if shape is not None:
@@ -357,6 +232,16 @@ class _compressed_sparse_matrix(sparse_data._data_matrix,
             if shape is None:
                 shape = arg1.shape
 
+        elif isinstance(arg1, tuple) and len(arg1) == 2:
+            # Note: This implementation is not efficeint, as it first
+            # constructs a sparse matrix with coo format, then converts it to
+            # compressed format.
+            sp_coo = coo.coo_matrix(arg1, shape=shape, dtype=dtype, copy=copy)
+            sp_compressed = sp_coo.asformat(self.format)
+            data = sp_compressed.data
+            indices = sp_compressed.indices
+            indptr = sp_compressed.indptr
+
         elif isinstance(arg1, tuple) and len(arg1) == 3:
             data, indices, indptr = arg1
             if not (base.isdense(data) and data.ndim == 1 and
@@ -913,87 +798,26 @@ class _compressed_sparse_matrix(sparse_data._data_matrix,
 
         Returns:
             (cupy.ndarray): Reduce result for nonzeros in each
-                major_index.
+            major_index.
 
         """
-
+        out_shape = self.shape[1 - axis]
         # Call to the appropriate kernel function
-        if axis == 1:
-            # Create the vector to hold output
-            value = cupy.zeros(self.shape[0]).astype(cupy.float64)
+        out = cupy.zeros(out_shape).astype(cupy.float64)
+        if nonzero:
+            kerns = {cupy.amax: self._max_nonzero_reduction_kern,
+                     cupy.amin: self._min_nonzero_reduction_kern}
+        else:
+            kerns = {cupy.amax: self._max_reduction_kern,
+                     cupy.amin: self._min_reduction_kern}
 
-            if nonzero:
-                # Perform the calculation
-                if ufunc == cupy.amax:
-                    self._max_nonzero_reduction_kern(
-                        (self.shape[0],), (1,),
-                        (self.data.astype(cupy.float64),
-                         self.indptr[:len(self.indptr) - 1],
-                         self.indptr[1:], cupy.int64(self.shape[1]),
-                         value))
-                if ufunc == cupy.amin:
-                    self._min_nonzero_reduction_kern(
-                        (self.shape[0],), (1,),
-                        (self.data.astype(cupy.float64),
-                         self.indptr[:len(self.indptr) - 1],
-                         self.indptr[1:], cupy.int64(self.shape[1]),
-                         value))
+        kerns[ufunc]((out_shape,), (1,),
+                     (self.data.astype(cupy.float64),
+                      self.indptr[:len(self.indptr) - 1],
+                      self.indptr[1:], cupy.int64(self.shape[axis]),
+                      out))
 
-            else:
-                # Perform the calculation
-                if ufunc == cupy.amax:
-                    self._max_reduction_kern(
-                        (self.shape[0],), (1,),
-                        (self.data.astype(cupy.float64),
-                         self.indptr[:len(self.indptr) - 1],
-                         self.indptr[1:], cupy.int64(self.shape[1]),
-                         value))
-                if ufunc == cupy.amin:
-                    self._min_reduction_kern(
-                        (self.shape[0],), (1,),
-                        (self.data.astype(cupy.float64),
-                         self.indptr[:len(self.indptr) - 1],
-                         self.indptr[1:], cupy.int64(self.shape[1]),
-                         value))
-
-        if axis == 0:
-            # Create the vector to hold output
-            value = cupy.zeros(self.shape[1]).astype(cupy.float64)
-
-            if nonzero:
-                # Perform the calculation
-                if ufunc == cupy.amax:
-                    self._max_nonzero_reduction_kern(
-                        (self.shape[1],), (1,),
-                        (self.data.astype(cupy.float64),
-                         self.indptr[:len(self.indptr) - 1],
-                         self.indptr[1:], cupy.int64(self.shape[0]),
-                         value))
-                if ufunc == cupy.amin:
-                    self._min_nonzero_reduction_kern(
-                        (self.shape[1],), (1,),
-                        (self.data.astype(cupy.float64),
-                         self.indptr[:len(self.indptr) - 1],
-                         self.indptr[1:], cupy.int64(self.shape[0]),
-                         value))
-            else:
-                # Perform the calculation
-                if ufunc == cupy.amax:
-                    self._max_reduction_kern(
-                        (self.shape[1],), (1,),
-                        (self.data.astype(cupy.float64),
-                         self.indptr[:len(self.indptr) - 1],
-                         self.indptr[1:], cupy.int64(self.shape[0]),
-                         value))
-                if ufunc == cupy.amin:
-                    self._min_reduction_kern(
-                        (self.shape[1],), (1,),
-                        (self.data.astype(cupy.float64),
-                         self.indptr[:len(self.indptr) - 1],
-                         self.indptr[1:], cupy.int64(self.shape[0]),
-                         value))
-
-        return value
+        return out
 
     def _arg_minor_reduce(self, ufunc, axis):
         """Reduce nonzeros with a ufunc over the minor axis when non-empty
@@ -1008,49 +832,31 @@ class _compressed_sparse_matrix(sparse_data._data_matrix,
 
         Returns:
             (cupy.ndarray): Reduce result for nonzeros in each
-                major_index
+            major_index
 
         """
 
         # Call to the appropriate kernel function
-        if axis == 1:
-            # Create the vector to hold output
-            value = cupy.zeros(self.shape[0]).astype(cupy.int64)
+        # Create the vector to hold output
+        # Note: it's important to set "int" here, following what SciPy
+        # does, as the outcome dtype is platform dependent
+        out_shape = self.shape[1 - axis]
+        out = cupy.zeros(out_shape, dtype=int)
 
-            # Perform the calculation
-            if ufunc == cupy.argmax:
-                self._max_arg_reduction_kern(
-                    (self.shape[0],), (1,),
-                    (self.data.astype(cupy.float64), self.indices,
-                     self.indptr[:len(self.indptr) - 1],
-                     self.indptr[1:], cupy.int64(self.shape[1]),
-                     value))
-            if ufunc == cupy.argmin:
-                self._min_arg_reduction_kern(
-                    (self.shape[0],), (1,),
-                    (self.data.astype(cupy.float64), self.indices,
-                     self.indptr[:len(self.indptr) - 1],
-                     self.indptr[1:], cupy.int64(self.shape[1]),
-                     value))
+        # Perform the calculation
+        ker_name = '_arg_reduction<{}, {}>'.format(
+            _scalar.get_typename(self.data.dtype),
+            _scalar.get_typename(out.dtype))
 
-        if axis == 0:
-            # Create the vector to hold output
-            value = cupy.zeros(self.shape[1]).astype(cupy.int64)
+        if ufunc == cupy.argmax:
+            ker = self._max_arg_reduction_mod.get_function('max' + ker_name)
+        elif ufunc == cupy.argmin:
+            ker = self._min_arg_reduction_mod.get_function('min' + ker_name)
 
-            # Perform the calculation
-            if ufunc == cupy.argmax:
-                self._max_arg_reduction_kern(
-                    (self.shape[1],), (1,),
-                    (self.data.astype(cupy.float64), self.indices,
-                     self.indptr[:len(self.indptr) - 1],
-                     self.indptr[1:], cupy.int64(self.shape[0]),
-                     value))
-            if ufunc == cupy.argmin:
-                self._min_arg_reduction_kern(
-                    (self.shape[1],), (1,),
-                    (self.data.astype(cupy.float64), self.indices,
-                     self.indptr[:len(self.indptr) - 1],
-                     self.indptr[1:],
-                     cupy.int64(self.shape[0]), value))
+        ker((out_shape,), (1,),
+            (self.data, self.indices,
+             self.indptr[:len(self.indptr) - 1],
+             self.indptr[1:], cupy.int64(self.shape[axis]),
+             out))
 
-        return value
+        return out
