@@ -1,18 +1,21 @@
-import unittest
+import math
 
 import numpy
 import pytest
 
 import cupy
-import cupy.core._accelerator as _acc
-from cupy.core import _cub_reduction
+import cupy._core._accelerator as _acc
+import cupy.cuda.cutensor
+from cupy._core import _cub_reduction
 from cupy import testing
 
 
 @testing.gpu
-class TestSumprod(unittest.TestCase):
+class TestSumprod:
 
+    @pytest.fixture(autouse=True)
     def tearDown(self):
+        yield
         # Free huge memory for slow test
         cupy.get_default_memory_pool().free_all_blocks()
         cupy.get_default_pinned_memory_pool().free_all_blocks()
@@ -164,7 +167,7 @@ class TestSumprod(unittest.TestCase):
     def test_sum_out_wrong_shape(self):
         a = testing.shaped_arange((2, 3, 4))
         b = cupy.empty((2, 3))
-        with self.assertRaises(ValueError):
+        with pytest.raises(ValueError):
             a.sum(axis=1, out=b)
 
     @testing.for_all_dtypes()
@@ -199,6 +202,11 @@ class TestSumprod(unittest.TestCase):
         a = testing.shaped_arange((2, 3), xp, src_dtype)
         return a.prod(dtype=dst_dtype)
 
+    @testing.numpy_cupy_allclose()
+    def test_product_alias(self, xp):
+        a = testing.shaped_arange((2, 3), xp, xp.float32)
+        return xp.product(a)
+
 
 # This class compares CUB results against NumPy's
 @testing.parameterize(*testing.product({
@@ -207,26 +215,27 @@ class TestSumprod(unittest.TestCase):
     'backend': ('device', 'block'),
 }))
 @testing.gpu
-@unittest.skipUnless(cupy.cuda.cub.available, 'The CUB routine is not enabled')
-class TestCubReduction(unittest.TestCase):
+@pytest.mark.skipif(
+    not cupy.cuda.cub.available, reason='The CUB routine is not enabled')
+class TestCubReduction:
 
+    @pytest.fixture(autouse=True)
     def setUp(self):
-        self.old_routine_accelerators = _acc.get_routine_accelerators()
-        self.old_reduction_accelerators = _acc.get_reduction_accelerators()
+        old_routine_accelerators = _acc.get_routine_accelerators()
+        old_reduction_accelerators = _acc.get_reduction_accelerators()
         if self.backend == 'device':
             _acc.set_routine_accelerators(['cub'])
             _acc.set_reduction_accelerators([])
         elif self.backend == 'block':
             _acc.set_routine_accelerators([])
             _acc.set_reduction_accelerators(['cub'])
-
-    def tearDown(self):
-        _acc.set_routine_accelerators(self.old_routine_accelerators)
-        _acc.set_reduction_accelerators(self.old_reduction_accelerators)
+        yield
+        _acc.set_routine_accelerators(old_routine_accelerators)
+        _acc.set_reduction_accelerators(old_reduction_accelerators)
 
     @testing.for_contiguous_axes()
     # sum supports less dtypes; don't test float16 as it's not as accurate?
-    @testing.for_dtypes('lLfdFD')
+    @testing.for_dtypes('qQfdFD')
     @testing.numpy_cupy_allclose(rtol=1E-5)
     def test_cub_sum(self, xp, dtype, axis):
         a = testing.shaped_random(self.shape, xp, dtype)
@@ -241,7 +250,7 @@ class TestCubReduction(unittest.TestCase):
         # xp is cupy, first ensure we really use CUB
         ret = cupy.empty(())  # Cython checks return type, need to fool it
         if self.backend == 'device':
-            func_name = 'cupy.core._routines_math.cub.'
+            func_name = 'cupy._core._routines_math.cub.'
             if len(axis) == len(self.shape):
                 func_name += 'device_reduce'
             else:
@@ -250,7 +259,7 @@ class TestCubReduction(unittest.TestCase):
                 a.sum(axis=axis)
         elif self.backend == 'block':
             # this is the only function we can mock; the rest is cdef'd
-            func_name = 'cupy.core._cub_reduction.'
+            func_name = 'cupy._core._cub_reduction.'
             func_name += '_SimpleCubReductionKernel_get_cached_function'
             func = _cub_reduction._SimpleCubReductionKernel_get_cached_function
             if len(axis) == len(self.shape):
@@ -264,7 +273,7 @@ class TestCubReduction(unittest.TestCase):
         return a.sum(axis=axis)
 
     # sum supports less dtypes; don't test float16 as it's not as accurate?
-    @testing.for_dtypes('lLfdFD')
+    @testing.for_dtypes('qQfdFD')
     @testing.numpy_cupy_allclose(rtol=1E-5, contiguous_check=False)
     def test_cub_sum_empty_axis(self, xp, dtype):
         a = testing.shaped_random(self.shape, xp, dtype)
@@ -276,7 +285,7 @@ class TestCubReduction(unittest.TestCase):
 
     @testing.for_contiguous_axes()
     # prod supports less dtypes; don't test float16 as it's not as accurate?
-    @testing.for_dtypes('lLfdFD')
+    @testing.for_dtypes('qQfdFD')
     @testing.numpy_cupy_allclose(rtol=1E-5)
     def test_cub_prod(self, xp, dtype, axis):
         a = testing.shaped_random(self.shape, xp, dtype)
@@ -291,7 +300,7 @@ class TestCubReduction(unittest.TestCase):
         # xp is cupy, first ensure we really use CUB
         ret = cupy.empty(())  # Cython checks return type, need to fool it
         if self.backend == 'device':
-            func_name = 'cupy.core._routines_math.cub.'
+            func_name = 'cupy._core._routines_math.cub.'
             if len(axis) == len(self.shape):
                 func_name += 'device_reduce'
             else:
@@ -300,7 +309,7 @@ class TestCubReduction(unittest.TestCase):
                 a.prod(axis=axis)
         elif self.backend == 'block':
             # this is the only function we can mock; the rest is cdef'd
-            func_name = 'cupy.core._cub_reduction.'
+            func_name = 'cupy._core._cub_reduction.'
             func_name += '_SimpleCubReductionKernel_get_cached_function'
             func = _cub_reduction._SimpleCubReductionKernel_get_cached_function
             if len(axis) == len(self.shape):
@@ -319,7 +328,7 @@ class TestCubReduction(unittest.TestCase):
     @testing.numpy_cupy_allclose(rtol=1E-4)
     def test_cub_cumsum(self, xp, dtype):
         if self.backend == 'block':
-            raise unittest.SkipTest('does not support')
+            pytest.skip('does not support')
 
         a = testing.shaped_random(self.shape, xp, dtype)
         if self.order in ('c', 'C'):
@@ -332,7 +341,7 @@ class TestCubReduction(unittest.TestCase):
 
         # xp is cupy, first ensure we really use CUB
         ret = cupy.empty(())  # Cython checks return type, need to fool it
-        func = 'cupy.core._routines_math.cub.device_scan'
+        func = 'cupy._core._routines_math.cub.device_scan'
         with testing.AssertFunctionIsCalled(func, return_value=ret):
             a.cumsum()
         # ...then perform the actual computation
@@ -344,7 +353,7 @@ class TestCubReduction(unittest.TestCase):
     @testing.numpy_cupy_allclose(rtol=1E-4)
     def test_cub_cumprod(self, xp, dtype):
         if self.backend == 'block':
-            raise unittest.SkipTest('does not support')
+            pytest.skip('does not support')
 
         a = testing.shaped_random(self.shape, xp, dtype)
         if self.order in ('c', 'C'):
@@ -358,7 +367,7 @@ class TestCubReduction(unittest.TestCase):
 
         # xp is cupy, first ensure we really use CUB
         ret = cupy.empty(())  # Cython checks return type, need to fool it
-        func = 'cupy.core._routines_math.cub.device_scan'
+        func = 'cupy._core._routines_math.cub.device_scan'
         with testing.AssertFunctionIsCalled(func, return_value=ret):
             a.cumprod()
         # ...then perform the actual computation
@@ -384,20 +393,21 @@ class TestCubReduction(unittest.TestCase):
     'order': ('C', 'F'),
 }))
 @testing.gpu
-@unittest.skipUnless(cupy.cuda.cutensor.available,
-                     'The cuTENSOR routine is not enabled')
-class TestCuTensorReduction(unittest.TestCase):
+@pytest.mark.skipif(
+    not cupy.cuda.cutensor.available,
+    reason='The cuTENSOR routine is not enabled')
+class TestCuTensorReduction:
 
+    @pytest.fixture(autouse=True)
     def setUp(self):
-        self.old_accelerators = cupy.core.get_routine_accelerators()
-        cupy.core.set_routine_accelerators(['cutensor'])
-
-    def tearDown(self):
-        cupy.core.set_routine_accelerators(self.old_accelerators)
+        old_accelerators = cupy._core.get_routine_accelerators()
+        cupy._core.set_routine_accelerators(['cutensor'])
+        yield
+        cupy._core.set_routine_accelerators(old_accelerators)
 
     @testing.for_contiguous_axes()
     # sum supports less dtypes; don't test float16 as it's not as accurate?
-    @testing.for_dtypes('lLfdFD')
+    @testing.for_dtypes('qQfdFD')
     @testing.numpy_cupy_allclose(rtol=1E-5, contiguous_check=False)
     def test_cutensor_sum(self, xp, dtype, axis):
         a = testing.shaped_random(self.shape, xp, dtype)
@@ -418,7 +428,7 @@ class TestCuTensorReduction(unittest.TestCase):
         return a.sum(axis=axis)
 
     # sum supports less dtypes; don't test float16 as it's not as accurate?
-    @testing.for_dtypes('lLfdFD')
+    @testing.for_dtypes('qQfdFD')
     @testing.numpy_cupy_allclose(rtol=1E-5, contiguous_check=False)
     def test_cutensor_sum_empty_axis(self, xp, dtype):
         a = testing.shaped_random(self.shape, xp, dtype)
@@ -439,7 +449,7 @@ class TestCuTensorReduction(unittest.TestCase):
     })
 )
 @testing.gpu
-class TestNansumNanprodLong(unittest.TestCase):
+class TestNansumNanprodLong:
 
     def _do_transposed_axis_test(self):
         return not self.transpose_axes and self.axis != 1
@@ -480,7 +490,7 @@ class TestNansumNanprodLong(unittest.TestCase):
     })
 )
 @testing.gpu
-class TestNansumNanprodExtra(unittest.TestCase):
+class TestNansumNanprodExtra:
 
     def test_nansum_axis_float16(self):
         # Note that the above test example overflows in float16. We use a
@@ -509,7 +519,7 @@ class TestNansumNanprodExtra(unittest.TestCase):
         a = testing.shaped_arange(self.shape)
         a[:, 1] = cupy.nan
         b = cupy.empty((2, 3))
-        with self.assertRaises(ValueError):
+        with pytest.raises(ValueError):
             cupy.nansum(a, axis=1, out=b)
 
 
@@ -520,7 +530,7 @@ class TestNansumNanprodExtra(unittest.TestCase):
     })
 )
 @testing.gpu
-class TestNansumNanprodAxes(unittest.TestCase):
+class TestNansumNanprodAxes:
     @testing.for_all_dtypes(no_bool=True, no_float16=True)
     @testing.numpy_cupy_allclose(rtol=1e-6)
     def test_nansum_axes(self, xp, dtype):
@@ -531,7 +541,7 @@ class TestNansumNanprodAxes(unittest.TestCase):
 
 
 @testing.gpu
-class TestNansumNanprodHuge(unittest.TestCase):
+class TestNansumNanprodHuge:
     def _test(self, xp, nan_slice):
         a = testing.shaped_random((2048, 1, 1024), xp, 'f')
         a[nan_slice] = xp.nan
@@ -556,20 +566,26 @@ axes = [0, 1, 2]
 
 @testing.parameterize(*testing.product({'axis': axes}))
 @testing.gpu
-class TestCumsum(unittest.TestCase):
+class TestCumsum:
+
+    def _cumsum(self, xp, a, *args, **kwargs):
+        b = a.copy()
+        res = xp.cumsum(a, *args, **kwargs)
+        testing.assert_array_equal(a, b)  # Check if input array is overwritten
+        return res
 
     @testing.for_all_dtypes()
     @testing.numpy_cupy_allclose()
     def test_cumsum(self, xp, dtype):
         a = testing.shaped_arange((5,), xp, dtype)
-        return xp.cumsum(a)
+        return self._cumsum(xp, a)
 
     @testing.for_all_dtypes()
     @testing.numpy_cupy_allclose()
     def test_cumsum_out(self, xp, dtype):
         a = testing.shaped_arange((5,), xp, dtype)
         out = xp.zeros((5,), dtype=dtype)
-        xp.cumsum(a, out=out)
+        self._cumsum(xp, a, out=out)
         return out
 
     @testing.for_all_dtypes()
@@ -577,21 +593,21 @@ class TestCumsum(unittest.TestCase):
     def test_cumsum_out_noncontiguous(self, xp, dtype):
         a = testing.shaped_arange((5,), xp, dtype)
         out = xp.zeros((10,), dtype=dtype)[::2]  # Non contiguous view
-        xp.cumsum(a, out=out)
+        self._cumsum(xp, a, out=out)
         return out
 
     @testing.for_all_dtypes()
     @testing.numpy_cupy_allclose()
     def test_cumsum_2dim(self, xp, dtype):
         a = testing.shaped_arange((4, 5), xp, dtype)
-        return xp.cumsum(a)
+        return self._cumsum(xp, a)
 
     @testing.for_all_dtypes()
     @testing.numpy_cupy_allclose(contiguous_check=False)
     def test_cumsum_axis(self, xp, dtype):
         n = len(axes)
         a = testing.shaped_arange(tuple(range(4, 4 + n)), xp, dtype)
-        return xp.cumsum(a, axis=self.axis)
+        return self._cumsum(xp, a, axis=self.axis)
 
     @testing.for_all_dtypes()
     @testing.numpy_cupy_allclose()
@@ -600,7 +616,7 @@ class TestCumsum(unittest.TestCase):
         shape = tuple(range(4, 4 + n))
         a = testing.shaped_arange(shape, xp, dtype)
         out = xp.zeros(shape, dtype=dtype)
-        xp.cumsum(a, axis=self.axis, out=out)
+        self._cumsum(xp, a, axis=self.axis, out=out)
         return out
 
     @testing.for_all_dtypes()
@@ -610,7 +626,7 @@ class TestCumsum(unittest.TestCase):
         shape = tuple(range(4, 4 + n))
         a = testing.shaped_arange(shape, xp, dtype)
         out = xp.zeros((8,)+shape[1:], dtype=dtype)[::2]  # Non contiguous view
-        xp.cumsum(a, axis=self.axis, out=out)
+        self._cumsum(xp, a, axis=self.axis, out=out)
         return out
 
     @testing.for_all_dtypes()
@@ -625,7 +641,7 @@ class TestCumsum(unittest.TestCase):
     def test_cumsum_axis_empty(self, xp, dtype):
         n = len(axes)
         a = testing.shaped_arange(tuple(range(0, n)), xp, dtype)
-        return xp.cumsum(a, axis=self.axis)
+        return self._cumsum(xp, a, axis=self.axis)
 
     @testing.for_all_dtypes()
     def test_invalid_axis_lower1(self, dtype):
@@ -637,7 +653,7 @@ class TestCumsum(unittest.TestCase):
     @testing.for_all_dtypes()
     def test_invalid_axis_lower2(self, dtype):
         a = testing.shaped_arange((4, 5), cupy, dtype)
-        with self.assertRaises(numpy.AxisError):
+        with pytest.raises(numpy.AxisError):
             return cupy.cumsum(a, axis=-a.ndim - 1)
 
     @testing.for_all_dtypes()
@@ -650,35 +666,41 @@ class TestCumsum(unittest.TestCase):
     @testing.for_all_dtypes()
     def test_invalid_axis_upper2(self, dtype):
         a = testing.shaped_arange((4, 5), cupy, dtype)
-        with self.assertRaises(numpy.AxisError):
+        with pytest.raises(numpy.AxisError):
             return cupy.cumsum(a, axis=a.ndim + 1)
 
     def test_cumsum_arraylike(self):
-        with self.assertRaises(TypeError):
+        with pytest.raises(TypeError):
             return cupy.cumsum((1, 2, 3))
 
     @testing.for_float_dtypes()
     def test_cumsum_numpy_array(self, dtype):
         a_numpy = numpy.arange(8, dtype=dtype)
-        with self.assertRaises(TypeError):
+        with pytest.raises(TypeError):
             return cupy.cumsum(a_numpy)
 
 
 @testing.gpu
-class TestCumprod(unittest.TestCase):
+class TestCumprod:
+
+    def _cumprod(self, xp, a, *args, **kwargs):
+        b = a.copy()
+        res = xp.cumprod(a, *args, **kwargs)
+        testing.assert_array_equal(a, b)  # Check if input array is overwritten
+        return res
 
     @testing.for_all_dtypes()
     @testing.numpy_cupy_allclose()
     def test_cumprod_1dim(self, xp, dtype):
         a = testing.shaped_arange((5,), xp, dtype)
-        return xp.cumprod(a)
+        return self._cumprod(xp, a)
 
     @testing.for_all_dtypes()
     @testing.numpy_cupy_allclose()
     def test_cumprod_out(self, xp, dtype):
         a = testing.shaped_arange((5,), xp, dtype)
         out = xp.zeros((5,), dtype=dtype)
-        xp.cumprod(a, out=out)
+        self._cumprod(xp, a, out=out)
         return out
 
     @testing.for_all_dtypes()
@@ -686,20 +708,20 @@ class TestCumprod(unittest.TestCase):
     def test_cumprod_out_noncontiguous(self, xp, dtype):
         a = testing.shaped_arange((5,), xp, dtype)
         out = xp.zeros((10,), dtype=dtype)[::2]  # Non contiguous view
-        xp.cumprod(a, out=out)
+        self._cumprod(xp, a, out=out)
         return out
 
     @testing.for_all_dtypes()
     @testing.numpy_cupy_allclose(rtol=1e-6)
     def test_cumprod_2dim_without_axis(self, xp, dtype):
         a = testing.shaped_arange((4, 5), xp, dtype)
-        return xp.cumprod(a)
+        return self._cumprod(xp, a)
 
     @testing.for_all_dtypes()
     @testing.numpy_cupy_allclose()
     def test_cumprod_2dim_with_axis(self, xp, dtype):
         a = testing.shaped_arange((4, 5), xp, dtype)
-        return xp.cumprod(a, axis=1)
+        return self._cumprod(xp, a, axis=1)
 
     @testing.for_all_dtypes()
     @testing.numpy_cupy_allclose()
@@ -744,22 +766,81 @@ class TestCumprod(unittest.TestCase):
     @testing.for_all_dtypes()
     def test_invalid_axis_upper2(self, dtype):
         a = testing.shaped_arange((4, 5), cupy, dtype)
-        with self.assertRaises(numpy.AxisError):
+        with pytest.raises(numpy.AxisError):
             return cupy.cumprod(a, axis=a.ndim)
 
     def test_cumprod_arraylike(self):
-        with self.assertRaises(TypeError):
+        with pytest.raises(TypeError):
             return cupy.cumprod((1, 2, 3))
 
     @testing.for_float_dtypes()
     def test_cumprod_numpy_array(self, dtype):
         a_numpy = numpy.arange(1, 6, dtype=dtype)
-        with self.assertRaises(TypeError):
+        with pytest.raises(TypeError):
             return cupy.cumprod(a_numpy)
 
+    @testing.numpy_cupy_allclose()
+    def test_cumproduct_alias(self, xp):
+        a = testing.shaped_arange((2, 3), xp, xp.float32)
+        return xp.cumproduct(a)
+
+
+@testing.parameterize(*testing.product({
+    'shape': [(20,), (7, 6), (3, 4, 5)],
+    'axis': [None, 0, 1, 2],
+    'func': ('nancumsum', 'nancumprod'),
+}))
+@testing.gpu
+class TestNanCumSumProd:
+
+    zero_density = 0.25
+
+    def _make_array(self, dtype):
+        dtype = numpy.dtype(dtype)
+        if dtype.char in 'efdFD':
+            r_dtype = dtype.char.lower()
+            a = testing.shaped_random(self.shape, numpy, dtype=r_dtype,
+                                      scale=1)
+            if dtype.char in 'FD':
+                ai = a
+                aj = testing.shaped_random(self.shape, numpy, dtype=r_dtype,
+                                           scale=1)
+                ai[ai < math.sqrt(self.zero_density)] = 0
+                aj[aj < math.sqrt(self.zero_density)] = 0
+                a = ai + 1j * aj
+            else:
+                a[a < self.zero_density] = 0
+            a = a / a
+        else:
+            a = testing.shaped_random(self.shape, numpy, dtype=dtype)
+        return a
+
+    @testing.for_all_dtypes()
+    @testing.numpy_cupy_allclose()
+    def test_nancumsumprod(self, xp, dtype):
+        if self.axis is not None and self.axis >= len(self.shape):
+            pytest.skip()
+        a = xp.array(self._make_array(dtype))
+        out = getattr(xp, self.func)(a, axis=self.axis)
+        return xp.ascontiguousarray(out)
+
+    @testing.for_all_dtypes()
+    @testing.numpy_cupy_allclose()
+    def test_nancumsumprod_out(self, xp, dtype):
+        dtype = numpy.dtype(dtype)
+        if self.axis is not None and self.axis >= len(self.shape):
+            pytest.skip()
+        if len(self.shape) > 1 and self.axis is None:
+            # Skip the cases where np.nancum{sum|prod} raise AssertionError.
+            pytest.skip()
+        a = xp.array(self._make_array(dtype))
+        out = xp.empty(self.shape, dtype=dtype)
+        getattr(xp, self.func)(a, axis=self.axis, out=out)
+        return xp.ascontiguousarray(out)
+
 
 @testing.gpu
-class TestDiff(unittest.TestCase):
+class TestDiff:
 
     @testing.for_all_dtypes()
     @testing.numpy_cupy_allclose()
@@ -822,3 +903,137 @@ class TestDiff(unittest.TestCase):
                 xp.diff(a, axis=3)
             with pytest.raises(numpy.AxisError):
                 xp.diff(a, axis=-4)
+
+
+# This class compares CUB results against NumPy's
+@testing.parameterize(*testing.product_dict(
+    testing.product({
+        'shape': [()],
+        'axis': [None, ()],
+        'spacing': [(), (1.2,)],
+    })
+    + testing.product({
+        'shape': [(33,)],
+        'axis': [None, 0, -1, (0,)],
+        'spacing': [(), (1.2,), 'sequence of int', 'arrays'],
+    })
+    + testing.product({
+        'shape': [(10, 20), (10, 20, 30)],
+        'axis': [None, 0, -1, (0, -1), (1, 0)],
+        'spacing': [(), (1.2,), 'sequence of int', 'arrays', 'mixed'],
+    }),
+    testing.product({
+        'edge_order': [1, 2],
+    }),
+))
+@testing.gpu
+class TestGradient:
+
+    def _gradient(self, xp, dtype, shape, spacing, axis, edge_order):
+        x = testing.shaped_random(shape, xp, dtype=dtype)
+        if axis is None:
+            normalized_axes = tuple(range(x.ndim))
+        else:
+            normalized_axes = axis
+            if not isinstance(normalized_axes, tuple):
+                normalized_axes = normalized_axes,
+            normalized_axes = tuple(ax % x.ndim for ax in normalized_axes)
+        if spacing == 'sequence of int':
+            # one scalar per axis
+            spacing = tuple((ax + 1) / x.ndim for ax in normalized_axes)
+        elif spacing == 'arrays':
+            # one array per axis
+            spacing = tuple(
+                xp.arange(x.shape[ax]) * (ax + 0.5) for ax in normalized_axes
+            )
+            # make at one of the arrays have non-constant spacing
+            spacing[-1][5:] *= 2.0
+        elif spacing == 'mixed':
+            # mixture of arrays and scalars
+            spacing = [xp.arange(x.shape[normalized_axes[0]])]
+            spacing = spacing + [0.5] * (len(normalized_axes) - 1)
+        return xp.gradient(x, *spacing, axis=axis, edge_order=edge_order)
+
+    @testing.for_dtypes('fFdD')
+    @testing.numpy_cupy_allclose(atol=1e-6, rtol=1e-5)
+    def test_gradient_floating(self, xp, dtype):
+        return self._gradient(xp, dtype, self.shape, self.spacing, self.axis,
+                              self.edge_order)
+
+    # unsigned int behavior fixed in 1.18.1
+    # https://github.com/numpy/numpy/issues/15207
+    @testing.with_requires('numpy>=1.18.1')
+    @testing.for_int_dtypes(no_bool=True)
+    @testing.numpy_cupy_allclose(atol=1e-6, rtol=1e-5)
+    def test_gradient_int(self, xp, dtype):
+        return self._gradient(xp, dtype, self.shape, self.spacing, self.axis,
+                              self.edge_order)
+
+    @testing.numpy_cupy_allclose(atol=2e-2, rtol=1e-3)
+    def test_gradient_float16(self, xp):
+        return self._gradient(xp, numpy.float16, self.shape, self.spacing,
+                              self.axis, self.edge_order)
+
+
+@testing.gpu
+class TestGradientErrors:
+
+    def test_gradient_invalid_spacings1(self):
+        # more spacings than axes
+        spacing = (1.0, 2.0, 3.0)
+        for xp in [numpy, cupy]:
+            x = testing.shaped_random((32, 16), xp)
+            with pytest.raises(TypeError):
+                xp.gradient(x, *spacing)
+
+    def test_gradient_invalid_spacings2(self):
+        # wrong length array in spacing
+        shape = (32, 16)
+        spacing = (15, cupy.arange(shape[1] + 1))
+        for xp in [numpy, cupy]:
+            x = testing.shaped_random(shape, xp)
+            with pytest.raises(ValueError):
+                xp.gradient(x, *spacing)
+
+    def test_gradient_invalid_spacings3(self):
+        # spacing array with ndim != 1
+        shape = (32, 16)
+        spacing = (15, cupy.arange(shape[0]).reshape(4, -1))
+        for xp in [numpy, cupy]:
+            x = testing.shaped_random(shape, xp)
+            with pytest.raises(ValueError):
+                xp.gradient(x, *spacing)
+
+    def test_gradient_invalid_edge_order1(self):
+        # unsupported edge order
+        shape = (32, 16)
+        for xp in [numpy, cupy]:
+            x = testing.shaped_random(shape, xp)
+            with pytest.raises(ValueError):
+                xp.gradient(x, edge_order=3)
+
+    def test_gradient_invalid_edge_order2(self):
+        # shape cannot be < edge_order
+        shape = (1, 16)
+        for xp in [numpy, cupy]:
+            x = testing.shaped_random(shape, xp)
+            with pytest.raises(ValueError):
+                xp.gradient(x, axis=0, edge_order=2)
+
+    @testing.with_requires('numpy>=1.16')
+    def test_gradient_invalid_axis(self):
+        # axis out of range
+        shape = (4, 16)
+        for xp in [numpy, cupy]:
+            x = testing.shaped_random(shape, xp)
+            for axis in [-3, 2]:
+                with pytest.raises(numpy.AxisError):
+                    xp.gradient(x, axis=axis)
+
+    def test_gradient_bool_input(self):
+        # axis out of range
+        shape = (4, 16)
+        for xp in [numpy, cupy]:
+            x = testing.shaped_random(shape, xp, dtype=numpy.bool_)
+            with pytest.raises(TypeError):
+                xp.gradient(x)
